// -- [ usi_8574a.c ] ---------------------------------------------------------
/*
* Ein ATTINY2313 emuliert einen PCF_8574 Portexpander.
* Das erste Datenbyte wird auf den Portpins des Controllers ausgegeben.
* Die weiteren Bytes knnen als Kommandos interpretiert werden, die irgendwelche Aktionen
* veranlassen. Oder sie knnen auf einem Portextender (Schieberegister) ausgegeben werden.
* Im TX-Buffer knnen Ergebnisse abgelegt werden, die sich die Master abholen muss.
* In diesem Beispiel werden die eingelesenen Bytes unverndert in den TX-Buffer geschrieben.
*
* Ausgabe: 	Byte.0 -> PinB.0				Shift-out 	PORT D
*				Byte.1 -> PinB.1				sCK			PD.0
*				Byte.2 -> PinB.2				SI				PD.1
*				Byte.3 -> PinB.3				RCK			PD.2
*				Byte.4 -> PinB.4
*				Byte.5 -> PinA.0
*				Byte.6 -> PinB.6
*				Byte.7 -> PinA.1
*
* Sollen weitere Ports geschaltet werden, dann knnen diese ber Schieberegister
* 'angeflanscht' werden.
* Die TWI-Adresse ist auf 0x50 (64) eingestellt und kann ber Jumper an
* PD.3 (A0), PD.4 (A1), PD.5 (A2) gendert werden (Reset erforderlich).
* Beim Start wird geprft, ob eine andere Adresse im Eeprom abgelegt ist:
* Die Adresse muss geradzahlig sein, 255 ist Merkmal fr unbeschriebenes Eeprom.
*
* 11.05.08	shift_out() gendert, getestet mit 74HC595
*           Der Sleep-Modus ist integriert.
* 02.09.08  das ndern der Basis-TWI-Adresse ber einen General-Call ist eingebaut
*           Es wird immer auf USI-START oder USI-STOP - Condition getestet
*
* Tabwidth: 3
*/
// ----------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <inttypes.h>
#include <util/delay.h>
#include "usi_twi_slave.h"

#define MAXINPUT 			8				// maximale Anzahl von Eingabebytes bei TWI-read
#define SLAVE_ADRESS 	64				// eigene TWI-Addresse

// Shift_out auf Schieberegister
#define SHIFT_PORT 		PORTD			// PINS zum Shiften
#define SHIFT_SCK 		PD0			// Clock
#define SHIFT_SI			PD1			// serial in (des Schieberegisters)
#define SHIFT_RCK			PD2			// storage register clock (
#define SHIFT_WAIT		5				// Wartezeit in usec fr 1/2 Takt
												// bei 4.000.000 -> 80 Takte/Byte -> 50.000Byte/sec
#define PORT_INVERTIEREN				// bei logisch "0" liegt der Pegel am Pin auf high
												// wie beim PCF8574 - aus der Sicht des mc invertiert !
//#define INCLUDE_SHIFT_OUT			// den Programmcode fr das Shiften einbinden
#define SLEEP							   // zum Aktivieren des SLEEP-Modus

// Globale Variable
//volatile uint8_t General_Call;
// Prototypen
void shift_out(uint8_t data[], uint8_t anzahl);


// ----------------------------------------------------------------------------
int main( void )
{
register uint8_t j, i;					// Zhler
uint8_t buffer[TWI_RX_BUFFER_SIZE];	// enthlt die Daten der Ports
// freie Ports: PB.0, PB.1, PB.2, PB.3, PB.4, PB.6
//              PD.0, PD.1, PD.2, PD.3, PD.4, PD.5, PD.6
//              PA.0, PA.1, (PA.2 - dann aber kein Reset !)
General_Call = FALSE;

#ifdef INCLUDE_SHIFT_OUT				// Shiftpins als Ausgang und auf High setzen
DDRD 	= (1<<PD2 | 1<<PD1 | 1<<PD0);	// zum Shiften PD.0-PD.2 als Ausgang
												// Adresslage PD.3 - PD.5 als Eingang
												// Pullups fr Adresspins setzen, Shiftpins auf High
PORTD = (1<<PD5 | 1<<PD4 | 1<<PD3 | 1<<PD2 | 1<<PD1 | 1<<PD0);
#else											// ohne Shiftpins
												// Adresslage PD.3 - PD.5 als Eingang
PORTD = 	(1<<PD5 | 1<<PD4 | 1<<PD3);// Pullups fr Adresspins setzen
#endif

#ifdef PORT_INVERTIEREN
	PORTA = (1<<PA1 | 1<<PA0);			// Pins auf high legen fr 0
#endif
DDRA 	= 	(1<<PA1 | 1<<PA0);			// fr die beiden benutzten Pins von PORTA
#ifdef PORT_INVERTIEREN
	PORTB = (1<<PB6 | 1<<PB4 | 1<<PB3 | 1<<PB2 | 1<< PB1 | 1<<PB0);// Pins auf high legen fr 0
#endif										// PORTS PB.7 = SCK, PB.5 = SDA
DDRB 	= 	(1<<PB6 | 1<<PB4 | 1<<PB3 | 1<<PB2 | 1<< PB1 | 1<<PB0);

												// Stromsparmodus , aktiviert direkt vor sleep
//MCUCR = (1<<SM1 | 1<< SM0 | 1<<SE);// Powerdown-Modus, Sleep enable

ACSR  = (1<<ACD);							// Analog-Comparator disable

sei();	 									// enable_interrupt

usi_twi_slave_init();					// den Slave und seine Adresse initialisieren
												// set USI_State = USI_WAIT

while(1)										// Endlosschleife
	{
	while (USI_State & 0xF0);			// warten, bis eine laufende Transaktion beendet ist

	if (USI_State)  						// wenn ein STOP oder NACK vorliegt (jedoch kein USI_WAIT!)
		{										// bei USI_WAIT wird direkt in den SLEEP-Modus gewechselt
		if (General_Call)
			{
			General_Call = FALSE;		// General Call Flag zurcksetzen
			eeprom_write_twi_adr();		// gewnschte Adresse prfen und ins Eeprom schreiben
			}

		else
			{
			j = usi_twi_rx_bytes(); 	// Anzahl der empfangenen Bytes holen
												// und den Puffer auslesen
			usi_twi_get_data(buffer, j);
												// die eingelesenen Daten auf Portpins ausgeben
												// beginnend mit PORTB

												// bei Invertieren beachten, dass nur die Datenbits
												// invertiert werden, nicht die anderen Portbits !!

// buffer[0] wird auf den freien Portpins von PORTB und PORTA ausgegeben

#ifdef PORT_INVERTIEREN					// Ausgabe der Datenbits invertiert
			PORTB = ((PORTB & (1<<PB7 | 1<< PB5)) | (~(buffer[0] & 0b01011111)));
#else
			PORTB = ((PORTB & (1<<PB7 | 1<< PB5)) | (buffer[0] & 0b01011111));
#endif
			i = 0;							// nun noch die Pins von PORTA ausgeben
			if (buffer[0] & 0x20) i = 1;
			if (buffer[0] & 0x80) i+= 2;
#ifdef PORT_INVERTIEREN
			PORTA = (( PORTA & 0b11111100) | (~i)); //Ausgabe der Datenbits invertiert
#else
			PORTA = (( PORTA & 0b11111100) | i); //Ausgabe normal
#endif

#ifdef INCLUDE_SHIFT_OUT
			if (j > 1) shift_out(buffer, j); // wenn mehr als 1 Byte empfangen,
													// dann den Rest auf Schieberegister ausgeben
#endif

// Hier werden irgendwelche Aktionen ausgefhrt - oder auch nicht
// z.B.
// die Daten ggf. aufbereiten und zurckschreiben in den Sendepuffer
// Zum Testen werden die empfangenen Bytes unverndert in den TX-Buffer geschrieben
			usi_twi_put_data(buffer, j);

// Flag setzen, damit der Ruhezustand
			USI_State = USI_WAIT;
			}									// if (General_Call = False)
		}										// if (USI_State) - es gibt kein else

#ifdef SLEEP
// Energiesparoption wird ausgelst, wenn auf den USI-Start gewartet wird,
// erkennbar daran, dass das USI_State == USI_WAIT == 0 ist.

	if (USI_State == USI_WAIT)
		{
		MCUCR = (1<<SM1 | 1<< SM0 | 1<<SE);	// Powerdown-Modus, Sleep enable
		asm volatile ("sleep");
		MCUCR = (0<<SM1 | 0<< SM0 | 0<<SE);	// disable
		}
#else

	while (USI_State);					// Warteschleife anstelle des Sleep-Modus

#endif

	}											// while(1)
}

#ifdef INCLUDE_SHIFT_OUT
// ----------------------------------------------------------------------------
// schreibt 'anzahl' Bytes aus dem Array 'data' seriell an ein Schieberegister
void shift_out(uint8_t data[], uint8_t anzahl)
{
register uint8_t i, j;
uint8_t buffer;
												// die Shift-Pins des Ports auf 0 setzen
SHIFT_PORT &= (~((1<<SHIFT_SCK) | (1<<SHIFT_SI) | (1<<SHIFT_RCK)));

for (j = 1; j < anzahl; j++)
	{
	buffer = data[j];						// damit das Array selbst nicht gendert wird

	for (i = 0; i < 8; i++)
		{										// Start, Datenbit setzen, SCK = 0, RCK = 0
		if (buffer & 0x01) 				// wenn Bit.0 gesetzt
			{
			SHIFT_PORT |= (1<<SHIFT_SI);// dann SI auf high
			}
		else
			{
			SHIFT_PORT &= ~(1<<SHIFT_SI);// sonst SI auf low
			}

		_delay_us(SHIFT_WAIT);			// einen halben Takt warten
												// Daten + Clock schreiben - hier werden die Daten bernommen
		SHIFT_PORT |= (1<<SHIFT_SCK);	// Clock auf high

		_delay_us(SHIFT_WAIT);			// einen halben Takt warten
												// (dieser Halbtakt wird aber lnger als der erste)

		SHIFT_PORT &= ~(1<<SHIFT_SCK);// Clock wieder auf low
		buffer = buffer >> 1;			// das Datenbyte um 1 nach rechts schieben
		}										// das Ganze nun noch 7x je Byte
	}
	SHIFT_PORT |= ((1<<SHIFT_SCK) | (1<<SHIFT_SI) | (1<<SHIFT_RCK));
												// Daten aus Storage-Register (im Schieberegister)
												// in den BUffer schreiben und damit ausgeben,
												// SCK und SI auf high
}
#endif

